/*
 *          	        forktty
 *
 *	This program forks, calls setpgrp(), and closes stdin, stdout and 
 *      stderr.  Then it tries to execute the command line.
 *
 *      Copying, modification, ... : Gnu Public License 
 */

#include <sys/types.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/wait.h>  /* WEXITSTATUS */
#include <signal.h>
#include <unistd.h>  /* setsid() */
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>

#ifndef VERBOSE
#define VERBOSE 0
#endif

#ifndef VERSION
#define VERSION "0"
#endif
char version[]= VERSION ", " __DATE__ ", " __TIME__ ;

#ifndef O_NOCTTY
#define O_NOCTTY 0
#endif


char starthost[256], progname[1025]={0};
char cname[]="forktty";

int verbose=VERBOSE;

static void usage(void ) {
  printf("usage: %s [-dhfnpVvw] [-l /dev/tty??] [-c command] program_to_run\n",
	 progname);
  printf("Try `%s -h' for more information \n", progname);
}

static void help(void ) {
  printf("usage: %s [-dhfnpVvw] [-l /dev/tty??] [-c command] program_to_run\n",
	 progname);
  printf("%s %s\n"
	 "  -c command     run child  bash -c command \n"
	 "  -d             login shell (prepends a '-') \n"
	 "  -f             fork only, no setsid() nor TIOCNOTTY \n"
	 "  --fdi <int>    dup2() stdin to the file descriptor <int>, also \n"
	 "  --fd[o,e] <i>  dup2() std[out,err] to fd <i> \n"
	 "  -l /dev/tty??  controlling terminal to open for child \n"
	 "  -l &<int>      use file descriptor <int> for child CTTY \n"
	 "  -n             don't initialize (stty TCGETS) terminal \n"
	 "  -p             pipe stdin and stdout to tty \n"
	 "  -r             use open(NONBLOCK) close() open() sequence \n"
	 "  -V             version \n"
	 "  -v             verbose \n"
	 "  -w             wait for program termination \n"
	 "  --sane  --baud <int>  --hupcl  --clocal  --crtscts  see stty(1)\n",
	 cname,version);
}

/* server */
#define MAX_ARGS 128
#define MAX_FLEN 1025
#define EXEC_PATH "/bin"
#define EXEC_SHELL "bash"

#define WHERE "server"
int server(int ttyfd, int ofork, int ologin, char *command, int argc, char **argv, int verbose) {
  char where[]=WHERE;
  int err=0, pgrp;
  int nargc=0;                 /* argc for execvp */
  char *nargv[MAX_ARGS+1];     /* rebuild the argument array */
  char *file_name;
  char file_login[MAX_FLEN+1]="-", opt_c[]="-c";
  char exec_path_name[]=EXEC_PATH "/" EXEC_SHELL;

  /*** prepare prog exec ***/
  if(command) {  /* -c */
    nargv[0]=exec_path_name;
    nargv[1]=opt_c;
    nargv[2]=command;
    nargc=3;
  } 
  if(argc>0) {   /* normal, just a program name */
    int i;
    if(argc+nargc>=MAX_ARGS) {
      fprintf(stderr,"%s: too many arguments %i",where,argc);
      err=5; goto end_error;
    }
    for(i=0; i<argc; i++,nargc++) nargv[nargc]=argv[i];
  } /* if argc>0 */

  if(nargc<=0) {
    nargc=1;
    nargv[0]=exec_path_name;
  } 
  nargv[nargc]=NULL;

  file_name=nargv[0];
  
  if(verbose) 
    printf("%s: nargc=%d file_name=\"%s\" \r\n",where,nargc,file_name);
  
  if(ologin) {                               /* prepend a '-' to nargv[0][] */
    long flen=strlen(nargv[0]);
    if(flen>=MAX_FLEN) {
      fprintf(stderr,"%s: file too long %li\n",where,flen);
      err=7; goto end_error;
    }
    strcpy(&file_login[1],nargv[0]);
    nargv[0]=file_login;
  }

  /*** new CTTY ***/
  /* Close stdin and stderr so that they will not
   * be kept open.  Stdout is assumed to have been
   * redirected to some logging file, or /dev/null.
   * From now on, the daemon will not report any
   * error messages.  
   */

  /* return the prompt to the user */
  if(!(freopen("/dev/null","r",stdin))) 
    fprintf(stderr,"%s: Error %i-%s closing stdin\n",where,errno,
	   strerror(errno));

#ifdef USE_TIOCNOTTY
  /* old method */
  if(verbose>1) printf("%s: Disconnecting from tty using TIOCNOTTY\n",where);

  /* disconnect from tty */
  { int fd;                                /* stdout may be redirected */
    fd = open("/dev/tty", O_RDWR|O_NOCTTY);
    if(fd<0) {
      fprintf(stderr,"%s: Error %i-%s while detatching from terminal\n",
	      where,errno,strerror(errno));
    } else {
      /* Redirect stdin, stdout, and stderr to /dev/null. */
      /* freopen("/dev/null", "r", stdin); */
      freopen("/dev/null", "w", stdout);
      freopen("/dev/null", "w", stderr);
      /* Disconnect from the controlling tty. */
      (void )ioctl(fd, TIOCNOTTY, NULL);
      close(fd);
    }
  }
#endif 
  /* set the program group to the process id 
   * daemon(0,0);     last resort BSD Linux 
   * setpgrp();       calls setsid() on solaris  
   * setpgrp(0,0);    ultrix Linux  calls setpgid(0,0) 
   * setpgid(0,0);    BSD4.2 SVr4 POSIX Linux  
   * setsid();        HP-UX10 POSIX SVr4 Linux 
   */

  if(!ofork) {
    if(verbose>1) printf("%s: Disconnecting from tty using setsid()\n",where);
    if((pgrp=setsid())<0) /* set new session, controlling terminal is lost */
      fprintf(stderr,"%s: Error setsid() %i-%s",where,errno,strerror(errno));

    if(verbose) printf("%s: pgrp %d\n",where,pgrp);
  }

  /* A new controlling tty can only be allocated if the prog is a session 
   * leader; if it's PID=PGID=SID 
   */
  if(ttyfd>=0) {
    int i;
    for(i=0; i<3; i++)       /* dup2 closes the old stdin, stdout, stderr */
      if(dup2(ttyfd, i)==-1) { 
	fprintf(stderr,"%s: Error dup2(ttyfd=%i,%i) %i-%s \n",where,ttyfd,i,
		errno, strerror(errno));
	err=10; goto end_error;
      }

    if(ttyfd>2 && close(ttyfd)) {
      printf("%s: Error %i-%s closing ttyfd=%i \n",where, 
	     errno,strerror(errno), ttyfd);
      ttyfd=-1;
      err=11; goto end_error;
    }
    /* setbuf(stdin, (char *) NULL); 
     * setbuf(stdout, (char *) NULL);
     * setbuf(stderr, (char *) NULL); 
     */

#ifdef USE_TIOCNOTTY
    err=ioctl(0, TIOCSCTTY, NULL);  /* set controlling tty */
    if(err<0)
      printf("%s: Error %i-%s TIOCSCTTY \"%s\"\n", where, 
	     errno,strerror(errno),ttyfile);
#endif
  } /* new TTY specified */

  /*** exec prog ***/
  if(execvp(file_name, nargv)) {
    fprintf(stderr,"%s: Error exec() %i-%s\n", where, 
	    errno, strerror(errno));
    err=20; goto end_error;
  }

  /* stdin stdout stderr stdd[] are closed by OS on exit of prog */
end_error:
  return(err);
}
#undef WHERE

#undef EXEC_SHELL 
#undef EXEC_PATH 
#undef MAX_ARGS
#undef MAX_FLEN

/* bucket brigade
 * 
 * Writing on a pty will read the same characters back if the tty
 * echo flag is enabled.
 *
 * Repeating  bufin=read(infd[i], buf, sizeof(buf));  after success 
 * requires O_NONBLOCK on infd[i].
 *
 * readfds must be reinitialized if an infd[] is written.
 *
 * EOF is reported on a tty if the pty closes.  Otherwise Input/output error.
 */
#define WHERE "bbrigade"
int bbrigade(int ina, int *outa, int inb, int *outb, int verbose) {
  int err=0, i;
  char where[]=WHERE;
  char buf[2048];
  int infd[2], outfd[2], ineof[2]={0,0}, outeof[2]={0,0};
  FILE *outfp[2];

  infd[0]=ina;     infd[1]=inb;
  outfd[0]=*outa;  outfd[1]=*outb;

  for(i=0; i<2; i++) {
    if(!(outfp[i]=fdopen(outfd[i], "w"))) {
      fprintf(stderr,"%s: Error %i-%s opening fd=%i\n", where,
	      errno,strerror(errno), outfd[i]);
      err=1; goto end_error;
    }
  }

  for(; !err && !(ineof[0] && ineof[1]); ) {
    fd_set readfds;
    int readfdlen, inready;

    FD_ZERO(&readfds);   /* need reinit if some infd[i]==outfd[j] */
    readfdlen=0;
    for(i=0; i<2; i++) 
      if(!ineof[i]) {
	FD_SET(infd[i], &readfds);
	if(readfdlen<=infd[i]) 
	  readfdlen=infd[i]+1;
      }

    inready=select(readfdlen, &readfds, NULL, NULL, NULL);
    if(inready<0) {
      fprintf(stderr,"%s: Error select() %i-%s\n",where,
	      errno,strerror(errno));
      err=2; goto end_error;
    }

    for(i=0; i<2; i++) 
      if(!ineof[i] && FD_ISSET(infd[i], &readfds)) {
	long bufin, bufout;
	
	bufin=read(infd[i], buf, sizeof(buf));
	if(bufin<0) {
	  if(verbose) 
	    fprintf(stderr,"%s: Error %i-%s read(fd=%i)  EOF ?\n",where,
		    errno,strerror(errno),infd[i]);
	  /* err=3; break; */
	  bufin=0;
	}
	if(verbose>1) {
	  if(!bufin) buf[0]='\0';
	  fprintf(stderr,"copying %3li char %i -> %i  0x%02X\n",bufin, 
		 infd[i],outfd[i], (unsigned )buf[0]);
	}

	bufout=write(outfd[i], buf, bufin);
	if(bufout<0) { /* werror */
	  fprintf(stderr,"%s: Error %i-%s write(fd=%i)\n", where, 
		  errno,strerror(errno),outfd[i]);
	  err=4; break;
	} 
	if(bufout<bufin) { /* short write */
	  fprintf(stderr,"%s: wrote %li/%li bytes\n", where, 
		  bufout,bufin);
	  err=5; break;
	} 
	if(fflush(outfp[i])) {
	  fprintf(stderr,"%s: Error %i-%s write(fd=%i)\n", where, 
		  errno,strerror(errno),outfd[i]);
	  err=6; break;
	}

	if(bufin==0) { /* EOF on input infd[i] */
	  int ii;

	  if(verbose) 
	    fprintf(stderr,"%s: EOF fd[%i]=%i\n",where,i,infd[i]);

	  outeof[i]=ineof[i]=1;
	  if(close(outfd[i])) { /* close output */
	    fprintf(stderr,"%s: Error %i-%s close(fd=%i)\n", where, 
		    errno,strerror(errno),outfd[i]);
	    err=7; break;
	  }
	  
	  for(ii=0; ii<2; ii++) /* signal EOF if closed output is input */
	    if(infd[ii]==outfd[i]) ineof[ii]=1;

	  for(ii=0; ii<2; ii++) /* extra: signal EOF if infd[i] is output */
	    if(outfd[ii]==infd[i]) ineof[ii]=1;

	  if(verbose>1) 
	    fprintf(stderr,"%s: EOF ineof[]={%i, %i}\n",where,
		    ineof[0],ineof[1]);

	} /* EOF */ 
      } /* if input */
  } /* forever */

end_error:
  /* no need to close fp, they are shared with the fd */

  if(outeof[0]) *outa=-1;
  if(outeof[1]) *outb=-1;

  return(err);
}
#undef WHERE

/************************ change tty terminal line settings ******************
 * Should set  -echo  on the tty if it is a modem, because
 * otherwise the modem will receive its own answers.
 *
 * Terminal emulators set the tty in raw mode.  They also set stdin and 
 * stdout in raw mode to obtain an 8bit clean pipe: 
 *  user - stdin/stdout - term.emul. - tty - modem
 *
 */
#define BAUDS 10U
#define WHERE "termecho"
int termecho(int ttyfd, struct termios *termsave, int sane, int hupcl, int clocal, int crtscts, long baud) {
  struct termios termset;
  char where[]=WHERE;
  int err=0;
  long unsigned iflags_clear= IGNBRK | PARMRK | INPCK | ISTRIP | INLCR 
    | IGNCR | IXANY | IXOFF;
  long unsigned iflags_set= IGNPAR;  /* IGNBRK for modem HUP ('\0' is read)*/
  long unsigned isane= BRKINT | ICRNL | IXON | IMAXBEL;
  long unsigned oflags_clear= OLCUC | OCRNL | ONOCR | ONLRET | OFILL | OFDEL;
  long unsigned oflags_set= 0; 
  long unsigned osane= OPOST  | ONLCR;
  long unsigned cflags_clear= CSTOPB;
  long unsigned cflags_set= CS8 | CREAD;
  long unsigned lflags_clear= ECHONL | NOFLSH | TOSTOP | ECHOPRT;
  long unsigned lflags_set= 0;
  long unsigned lsane=ISIG | ICANON | ECHO | ECHOE | ECHOK | ECHOCTL 
    | ECHOKE | IEXTEN;

  err=ioctl(ttyfd, TCGETS, &termset);
  if(err) {
    fprintf(stderr,"%s: Error %i-%s ioctl TCGETS\n",where,
	    errno,strerror(errno));
    err=1; goto end_error;
  }
  *termsave=termset;

  if(sane) {
    iflags_set  |= isane;
    oflags_set  |= osane;
    lflags_set  |= lsane;
  } else {
    iflags_clear|= isane;
    oflags_clear|= osane;
    lflags_clear|= lsane;
  }

  termset.c_iflag &= ~iflags_clear;
  termset.c_iflag |=  iflags_set;

  termset.c_oflag &= ~oflags_clear;
  termset.c_oflag |=  oflags_set;

  if(baud>=0) {
    int bauds=BAUDS, i;
    long unsigned baudcflags[BAUDS]={
      B0, B300, B600, B1200, B1800, B2400, B4800, B9600, B19200, B38400
    };
    long baudlist[BAUDS]={
      0, 300, 600, 1200, 1800, 2400, 4800, 9600, 19200, 38400
    };
    
    for(i=0; i<bauds && baudlist[i]!=baud; i++);
    
    if(i>=bauds) {
      fprintf(stderr,"%s: baud=%li not found\n",where,baud);
      err=2; goto end_error;
    }
    cflags_clear|= CBAUD;
    cflags_set|= baudcflags[i];
  }

  if(hupcl) cflags_set|=HUPCL;
  else cflags_clear|=HUPCL;
  
  if(clocal) cflags_set|=CLOCAL;
  else cflags_clear|=CLOCAL;

  if(crtscts) cflags_set|=CRTSCTS;
  else cflags_clear|=CRTSCTS;
  
  termset.c_cflag &= ~cflags_clear;
  termset.c_cflag |=  cflags_set;

  termset.c_lflag &= ~lflags_clear;
  termset.c_lflag |=  lflags_set;

  if(sane) {
    int cind[]={VINTR, VEOF, VSTART, VSTOP, VSUSP, -1}, i;
    cc_t cdef[]={0x03, 0x04,   0x11,  0x13,  0x1A,  0};
    for(i=0; cind[i]>=0; i++) 
      if(!termset.c_cc[cind[i]]) 
	termset.c_cc[cind[i]]= cdef[i];
  }

  err=ioctl(ttyfd, TCSETSW, &termset);
  if(err) {
    fprintf(stderr,"%s: Error %i-%s ioctl TSETSW\n", where,
	    errno,strerror(errno));
    err=3; goto end_error;
  }

end_error:
  return(err);
}
#undef WHERE
#undef BAUDS

/*************************** main **************************************
 *
 *	This routine starts the server.  It forks, leaving the child
 *	to do all the work, so it does not have to be run in the
 *	background.  
 */
#define WHERE "main"
int main(int argc, char **argv) {
  char *where=cname; /* []=WHERE; */
  int err=0;
  pid_t pid;
  char *ttyfile=NULL, *command=NULL, ttybuf[257];
  int ttyfd=-1, ttyclose=0;
  int stdfd[3]={fileno(stdin), fileno(stdout), fileno(stderr)};
  int stdfdx[3]={-1, -1, -1};
  struct termios ttyios;
  int ologin=0, opipe=0, owait=0, onostty=0, oreopen=0, ofork=0,
    osane=0, ohupcl=0, oclocal=1, ocrtscts=1;
  long obaud=-1;

  extern char *optarg;
  extern int optind;
  int opt;
  int option_index; /* long opt */
  struct option long_options[]={ 
    {"sane",0,0,0}, {"hupcl",0,0,0}, {"clocal",0,0,0}, {"crtscts",0,0,0}, 
    {"baud",1,0,0}, {"fdi",1,0,0},   {"fdo",1,0,0},    {"fde",1,0,0},
    {0,0,0,0} 
  };

  
  /* fill starthost variable */
  if(gethostname(starthost, sizeof(starthost))) 
    strcpy(starthost,"nohostname");

  if(argc) {
    strncpy(progname,argv[0], sizeof(progname)-1);
    progname[sizeof(progname)-1]=0;
  }
  if(argc<=1) {
    usage();
    err=1; goto end_error;
  }

  optarg=NULL;
  optind=0;

  while((opt=getopt_long(argc, argv, "c:dhfl:nprVvw", 
			 long_options,&option_index)) != EOF) {
    int this_optind=optind-1;
    switch(opt) {
      int nums;
    case 0:          /* long option  option_index  is valid */
      switch(option_index) {
      case 0: osane^=1; break;
      case 1: ohupcl^=1; break;
      case 2: oclocal^=1; break;
      case 3: ocrtscts^=1; break;
      case 4: 
	if((nums=sscanf(optarg,"%li",&obaud))!=1) {
	  fprintf(stderr,"%s: error %i reading integer from \"%.10s\" \n", 
		  where, nums,optarg); 
	  err=2; goto end_error;
	}
	break;
      case 5:
      case 6:
      case 7:
	{ int newfd, num;
  	  if((num=sscanf(optarg,"%i",&newfd))!=1) {
	    fprintf(stderr,"%s: error %i reading integer from \"%.10s\" \n", 
		    where, num,optarg); 
	    err=3; goto end_error;
	  }
	  stdfdx[option_index-5]=newfd;
	}
	break;
      default:
	fprintf(stderr,"%s: ?? getopt returned option_index %d \n", 
		     where,option_index);
	err=4; goto end_error;
      }
      break;
    case 'c':
      command=optarg;
      if(!strlen(command)) {
	fprintf(stderr, "%s: zero optarg string length, option %i \n", where,
		this_optind);
	err=1; goto end_error;
      }
      break;
    case 'd':
      ologin^=1;
      break;
    case 'h':
      help();
      err=1; break;
    case 'f':
      ofork^=1;
      break;
    case 'l':
      ttyfile=optarg;
      break;
    case 'n':
      onostty^=1;
      break;
    case 'p':
      opipe^=1;
      break;
    case 'r':
      oreopen^=1;
      break;
    case 'V':
      printf("%s: Version %s\n",where,version);
      err=2; break;
    case 'v':
      verbose++;
      break;
    case 'w':
      owait^=1;
      break;
    case '?':
      fprintf(stderr,"%s: Getopt returned error at option %i\n",
	      where,this_optind);
      err=11; break;
    default:
      fprintf(stderr,"%s: ?? getopt returned character code 0x%02X\n", 
	      where,opt);
      err=12; break;
    }
    if(err) goto end_error;
  }

  /************* process options ***********/
  if(verbose) {
    if(verbose>1) {
      int flags=0xFFFFFFFF; 
      flags=fcntl(stdfd[0],F_GETFL);
      printf("%s: stdin fd=%i flags <", where,stdfd[0]);
      if (flags & O_RDONLY) printf("readonly "); 
      if (flags & O_WRONLY) printf("writeonly "); 
      if (flags & O_NDELAY) printf("nonblocking "); 
      if (flags & O_SYNC) printf("synchronous ");
      printf(">\n");
    }
    printf("%s: started @%s \n",where,starthost);
  }

  /* copy stdin  --fdi --fdo --fde */
  { int i;
    for(i=0; i<3; i++) 
      if(stdfdx[i]>=0) {
	if(dup2(stdfd[i],stdfdx[i])<0) {
	  fprintf(stderr,"%s: Error %i-%s dup2(fd=%i,fdx=%i) \n", where,
		  errno,strerror(errno), stdfd[i],stdfdx[i]);
	  err=15; goto end_error;      
	}
	if(verbose>1)
	  printf("%s: dup2(%i,%i) \n", where, stdfd[i],stdfdx[i]);
      } /* if >=0 */
  }

  if(opipe && !ttyfile) {
    fprintf(stderr,"%s: Error no ttyfile specified\n", where);
    err=16; goto end_error;
  }
  if(ttyfile) {
    /* bucket brigade: open the tty now to avoid HUPCL on prog exit */
    int ttylen=strlen(ttyfile);
    if(ttylen>1 && ttyfile[0]=='&') {
      int num;
      if((num=sscanf(&ttyfile[1], "%i",&ttyfd))!=1 || ttyfd<0) {
	fprintf(stderr,"%s: error %i reading integer from \"%.10s\" \n", 
		where, num,&ttyfile[1]); 
	err=17; goto end_error;
      }
    } else {
      if(ttylen && ttyfile[0]!='/') {      /* check for pathname */
	snprintf(ttybuf,sizeof(ttybuf)-1, "/dev/%s",ttyfile);
	ttybuf[sizeof(ttybuf)-1]='\0';
	ttyfile=ttybuf;
      }

      /* XXX should try to lock the line */
      if(!onostty && oreopen)
	ttyfd=open(ttyfile, O_RDWR | O_NONBLOCK);
      else
	ttyfd=open(ttyfile, O_RDWR);         /* blocks if !clocal && !DSR */
      if(ttyfd<0) {
	fprintf(stderr,"%s: Error %i-%s opening \"%s\" \n",
		where,errno,strerror(errno), ttyfile);
	err=18; goto end_error;
      }
      ttyclose=1;
    } /* if ttyfile[0]=='&' */
    if(verbose>1) printf("%s: open(%s) fd=%i \n", where, ttyfile,ttyfd);

    if(!onostty) {
      if((err=termecho(ttyfd, &ttyios, osane,ohupcl,oclocal,ocrtscts,obaud))) {
	fprintf(stderr,"%s: Error %i setting -echo on fd=%i \"%s\"\n", where,
		err,ttyfd,ttyfile);
	err=19; goto end_error;      
      }
      if(verbose>1) 
	printf("%s: TCSETSW fd=%i sane=%i hupcl=%i clocal=%i crtsctr=%i "
	       "baud=%li\n", where, ttyfd, osane, ohupcl, oclocal, ocrtscts, 
	       obaud);
      if(ttyclose && oreopen) {
	ttyclose=0;
	if(close(ttyfd)) {
	  fprintf(stderr,"%s: Error %i-%s ttyfd close(fd=%i) \n", where, 
		  errno,strerror(errno),ttyfd);
	  err=20; goto end_error;
	}

	/* XXX remove my lock */
	ttyfd=open(ttyfile, O_RDWR);         /* blocks if !clocal && !DSR */
	if(ttyfd<0) {
	  fprintf(stderr,"%s: Error reopen(%s): %i-%s \n",
		  where, ttyfile, errno,strerror(errno));
	  err=21; goto end_error;
	}
	ttyclose=1;
	/* XXX try to lock the line */
	if(verbose>1) printf("%s: reopen(%s) fd=%i \n", where, ttyfile,ttyfd);
      }
    } /* if !onostty */
  } /* if ttyfile */

  /* Set SIGCLD to SIG_IGN, in order to prevent
   * the accumulation of zombies as each child
   * terminates.  This means the running program does not
   * have to make wait calls to clean them up, and can safely 
   * exit.  Not allowed by POSIX.
   */
  if(!owait) signal(SIGCLD, SIG_IGN);
  
  fflush(stdout);  /* avoid double messages due to buffering */

  /* We could do a setpgrp() here.  The parent will be a group leader,
   * but the child not.  The child loses its controlling terminal at the
   * moment that the parent quits.  If the child were to open a terminal, it
   * will not be associated with that terminal as its control terminal since
   * it is not a group leader.
   */

  if(argc-optind>0 || command) { /* program name given */
    switch (pid=fork()) {
    case -1:		/* Unable to fork, for some reason. */
      fprintf(stderr, "%s: Error %i-%s unable to fork daemon\n", where,
	      errno,strerror(errno));
      err=15; goto end_error;
    case 0:		/* The child process (daemon) comes here. */
      err=server(ttyfd, ofork, ologin, command, argc-optind, &argv[optind], 
		 verbose);
      exit(err);
    default:		/* Parent process comes here. */
      if(verbose) printf("%s: childpid %d\n",where,pid);
    }

    if(owait) {
      int status, errchild;
      pid_t pidexit;

      if(verbose>1) printf("%s: waiting for child %i\n", where,pid);
      
      pidexit=waitpid(pid, &status, 0);
      if(pidexit<=0) {
	fprintf(stderr,"%s: Error waitpid() %i-%s pidexit=%i\n", where, 
		errno,strerror(errno),pidexit);
	err=20; goto end_error;
      }
      if(!WIFEXITED(status)) { /* uncought signal */
	fprintf(stderr,"%s: Child abnormal exit waitpid() pidexit=%i\n",
		where,pidexit);
	err=21; goto end_error;
      }

      if((errchild=WEXITSTATUS(status))) {
	fprintf(stderr,"%s: child status %i\n", where,errchild);
	err=64+errchild;
	goto end_error;
      }
    } /* waitopt */
  } /* if(argc-optind>0 || command) */

  if(opipe) {
    if((err=bbrigade(stdfd[0], &ttyfd, ttyfd, &stdfd[1], verbose))) {
      fprintf(stderr,"%s: Error %i bbrigade()\n", where,err);
      err=22; goto end_error;
    }
  }    

end_error:
  { int i;
    for(i=0; i<3; i++) 
      if(stdfdx[i]>=0 && close(stdfdx[i])) {
	fprintf(stderr,"%s: Error %i-%s close(fdx=%i) \n", where,
		errno,strerror(errno),stdfdx[i]);
	if(!err) err=31+i; 
      }
  }

  if(ttyfd>=0 && close(ttyfd)) {
    fprintf(stderr,"%s: Error %i-%s ttyfd close(fd=%i) \n", where, 
	    errno,strerror(errno),ttyfd);
    if(!err) err=98;
  }

  if(verbose>1) printf("%s: exiting err=%i\n", where,err);

  return(err);
}
#undef WHERE


