/* Copyright 2000 Servaas Vandenberghe, Leuven, Belgium, All rights reserved.
 This software is licensed under the GNU GENERAL PUBLIC LICENSE, version 2, or
 a later version at your option.

 Redistribution and use of this source code, with or without modification, is
 permitted provided that the following condition is met:

 1. Redistributions of this source code must retain the above copyright
    notice, this condition, and the following disclaimer.

  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
  EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>   /* getopt */
#include <stdio.h>
#include <string.h>
#include <errno.h> 

#ifndef VERBOSE
#define VERBOSE 0
#endif

#ifndef VERSION
#define VERSION "0.0"
#endif
static char version[]= VERSION ", " __DATE__ ", " __TIME__ ;

static char progname[1025]="getdsr";
static char cname[]="getdsr";

static void usage(void ) {
  printf("usage: %s [-DdhvV] [-s <int>] tty prog args \n", progname);
  printf("Try `%s -h' for more information \n", progname);
}

static void help(void) {
  printf("usage: %s [-DdhVv] [-s <int>] tty prog args  \n", progname);
  printf("%s %s \n", cname, version);
  printf("%s waits for a high DSR and executes prog \n",cname);
  printf("  -D        disable the DTR line \n"
	 "  -d        enable the DTR line \n"
	 "  -s <int>  sleep time before rechecking the DSR line \n"
	 "  -V        version   \n"
	 "  -v        verbose   \n"
	 );
}

/*
 *			M A I N
 *
 *	This routine opens the TTY and reads TIOCMGET.
 *
 */
#define DEBUG 0
#define WHERE "main"
int main(int argc, char *argv[]) {
  char *where=cname; /* []=WHERE; */
  char *ttyfile=NULL, buf[257];
  int err=0, verbose=0, ttyfd=0, ttyclose=0, ttycm=-1, ttysleep=30;
  int odtrs=0, odtrc=0;
  int opt;
  extern char *optarg;   /* pointer to the argument of the option */
  extern int optind;     /*, opterr, optopt; */

  if(argc) {
    strncpy(progname,argv[0], sizeof(progname)-1);
    progname[sizeof(progname)-1]=0;
  }
  if(argc<=1) {
    usage();
    err=1; goto end_error;
  }

  optarg=NULL;
  optind=0;

  while( (opt=getopt(argc, argv, "dDhl:s:vV")) != EOF ) {
    int this_optind=optind-1;
    switch(opt) {
    case 'D':
      odtrc^=1;
      break;
    case 'd':
      odtrs^=1;
      break;
    case 'h':
      help(); goto end_error;
    case 'l':
      ttyfile=optarg;
      break;
    case 's':
      { int errs=sscanf(optarg,"%i",&ttysleep);
        if(errs!=1) {
	  fprintf(stderr,"%s: Error %i: failed to extract an integer"
		  " from %.99s", where,errs,optarg);
	  err=10; goto end_error;
	}
      }
      break;
    case 'v':
      verbose++; break;
    case 'V':
      printf("%s %s\n", cname,version); goto end_error;
    case ':':
      fprintf(stderr,"%s: Expected parameter at option %i\n",where,
	      this_optind);
      err=10; goto end_error;	
    case '?':
      fprintf(stderr,"%s: Getopt returned error at option %i\n",
	      where,this_optind);
      err=11; goto end_error;
      break;
    default:
      fprintf(stderr,"%s: ?? getopt returned character code 0x%x\n",where,opt);
      err=12; goto end_error;
    }
  }

  /* check for ttyfile */
  if(!ttyfile && optind<argc) 
    ttyfile=argv[optind++];
  if(!ttyfile) {
    help(); goto end_error;
  }
  /* and progname */
  if(optind>=argc) {
    fprintf(stderr,"%s: program to execute not specified \n",where);
    err=31; goto end_error;
  }

  /* open the tty */
  { int ttylen=strlen(ttyfile);
    if(ttylen && (ttyfile[0]!='-' || ttylen>1)) {
      int busys=0;
      if(ttyfile[0]!='/') {      /* check for pathname */
	snprintf(buf,sizeof(buf)-1, "/dev/%s",ttyfile);
	buf[sizeof(buf)-1]='\0';
	ttyfile=buf;
      }
      while((ttyfd=open(ttyfile, O_NONBLOCK))<0) {
	busys++;
	if((errno!=EAGAIN && errno!=EBUSY) || busys>=3) {
	  fprintf(stderr,"%s: Error open(%s): %i-%s\n",where,
		  ttyfile,errno,strerror(errno));
	  err=29; goto end_error;
	}
	if(verbose) printf("%s: Retry open(%s): %i-%s\n",where,
			   ttyfile, errno,strerror(errno));
	sleep(30);
      } /* while */
      ttyclose=1;
      if(verbose>1) printf("%s: open(%s) fd=%i \n", where, ttyfile,ttyfd);
    } /* if tty not stdin */
  }

  /* clear and/or set DTR line.  Most OS systems set DTR on open() */
  if(odtrc || odtrs) {
    int i;
    if((err=ioctl(ttyfd, TIOCMGET, &ttycm))) {
      fprintf(stderr,"%s: Error ioctl(%i,TIOCMGET): %i-%s\n",where, ttyfd,
	      errno,strerror(errno));
      err=30; goto end_error;
    }
    for(i=0; i<2; i++) {
      if(!i) {
	if(odtrc) ttycm &= ~TIOCM_DTR; else continue;
      } else {
	if(odtrs) ttycm |=  TIOCM_DTR; else continue;
      }
      if((err=ioctl(ttyfd, TIOCMSET, &ttycm))) {
	fprintf(stderr,"%s: Error ioctl(%i,TIOCMSET): %i-%s\n",where, ttyfd,
		errno,strerror(errno));
	err=31; goto end_error;
      }
      sleep(1); 
    } /* for i */
  } else /* if DTR line */
    if(ttyclose) sleep(1); 

  /* get the modem status lines 
   * This is a NOOP for tty/pty pairs 
   */
  do {
    (void) ioctl(ttyfd, TIOCMGET, &ttycm);
    if(verbose) {
      int listcm[]={TIOCM_LE, TIOCM_DTR, TIOCM_RTS, TIOCM_ST, TIOCM_SR, 
		    TIOCM_CTS, TIOCM_CAR, TIOCM_RNG, TIOCM_DSR};
      char *listmsg[]={"line enable","data terminal ready","request to send",
		       "secondary transmit","secondary receive",
		       "clear to send","carrier detect","ring",
		       "data set ready", NULL};
      int i;
      for(i=0; listmsg[i]; i++) {
	if(ttycm & listcm[i]) printf("1 %s\n",listmsg[i]); 
	else printf("0 %s\n",listmsg[i]);
      }
    } /* if verbose */
    if(!(ttycm & TIOCM_DSR)) {
      if(!ttysleep) {
	err=35;
	if(verbose) printf("%s: aborting with return code %i \n",where,err);
	goto end_error;
      }
      if(verbose) printf("%s: sleeping %i s \n", where,ttysleep);
      sleep(ttysleep);
    } /* if DSR low */
  } while(!(ttycm & TIOCM_DSR));

  /* execute the program */
  if(ttyclose) {
    if(close(ttyfd)) {
      fprintf(stderr,"%s: close: %i-%s\n",where,errno,strerror(errno));
      err=40; goto end_error;
    }
    ttyclose=0;
  }
  { int filelen=strlen(argv[optind]);
    if(filelen && (argv[optind][0]!='-' || filelen>1)) {
      if(execvp(argv[optind], &argv[optind])) {
	fprintf(stderr,"%s: Error execvp: %i-%s", where,
		errno,strerror(errno));
	err=41; goto end_error;
      }
    } /* if filelen */
  }

end_error:
  if(ttyclose) 
    if(close(ttyfd)) {
      fprintf(stderr,"%s: Error close: %i-%s\n", where, errno,strerror(errno));
      if(!err) err=98;
    }

  return(err);
}
#undef WHERE
#undef DEBUG

#undef VERBOSE
#undef VERSION
